# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-22 11:20
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0005_auto_20160411_0250'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='additional_data',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='data',
            field=django.contrib.postgres.fields.jsonb.JSONField(),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='additional_data',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='data',
            field=django.contrib.postgres.fields.jsonb.JSONField(),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import logging
from datetime import timedelta

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models, connection
from django.db.models.query import QuerySet
from django.utils import six
from django.utils.timezone import now

str = unicode if six.PY2 else str

registered_models = {}

logger = logging.getLogger(__name__)


class HistoricalRecordQuerySet(QuerySet):
    def by_model_and_model_id(self, model, model_id):
        """
        Gets historical records by model and model id, so, basically the
        historical records for an instance of a model.
        :param model: Model which has the HistoricalRecord field.
        :param model_id: The model id for which you wish to get the history.
        :return: The instance's history.
        :rtype HistoricalRecord
        """
        return self.by_model(model).filter(object_id=model_id)

    def by_model(self, model):
        # noinspection PyProtectedMember
        """
        Gets historical records by model.
        :param model: Model which has the HistoricalRecord field.
        :return: The entire model's history.
        :rtype HistoricalRecord
        """
        return self.filter(
            content_type__model=model._meta.model_name,
            content_type__app_label=model._meta.app_label
        )

    def most_recent(self):
        """
        Gets the most recent historical record added to the database.
        :return: The most recent historical record added to the database.
        :rtype HistoricalRecord
        """
        return self.first()

    def older_than(self, days=None, weeks=None):
        """
        Gets all historical record entries that are older than either the
        number of days or the number of weeks passed.
        The weeks parameter will be preferred if both are supplied.
        :param days: Number of days old a historical record can be, at most.
        :param weeks: Number of weeks old a historical record can be, at most.
        :return: All the historical record entries that are older than the
                 given param.
        :rtype list(HistoricalRecord)
        """
        if not (days or weeks):
            logger.error('You must supply either the days or the weeks param')
            return
        elif days and weeks:
            logger.info('You supplied both days and weeks, weeks param'
                        ' will be used as the delimiter.')
        td = timedelta(weeks=weeks) if weeks else timedelta(days=days)
        return self.filter(history_date__lte=now() - td)

    def previous_version_by_model_and_id(self, model, object_id, history_id):
        """
        Returns the second to last snapshot of the history for model and
        instance id that is given.
        :param history_id: Id for history snapshot.
        :param model: The model for which the snapshot is for.
        :param object_id: The model ID for which the snapshot is for.
        :return: The previous to HistoricalRecord

        """
        main_qs = self.filter(
            content_type__model=model.model,
            content_type__app_label=model.app_label,
            object_id=object_id,
            id__lt=history_id
        )
        if main_qs.count() <= 0:
            return None
        result = main_qs.order_by('-history_date').first()

        return result

    def approx_count(self):
        """
            Takes a queryset and generates a fast approximate count(*) for it.
            This is required because postgresql count(*) has to go through all
            of the entries in the database, making it extremely slow for
            large tables.
            :return: int representing approx count(*)
            """
        table_name = self.model._meta.db_table
        cursor = connection.cursor()
        cursor.execute(
            "SELECT reltuples FROM pg_class WHERE relname='{}';".format(
                table_name))
        row = cursor.fetchone()
        return int(row[0])


class AbstractHistoricalRecord(models.Model):
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    history_date = models.DateTimeField(auto_now_add=True, db_index=True)
    history_user = models.CharField(max_length=50, null=True)
    history_user_id = models.PositiveIntegerField(null=True)
    history_type = models.CharField(
        max_length=1,
        choices=(
            ('+', 'Create'),
            ('~', 'Update'),
            ('-', 'Delete'),
        ),
        db_index=True
    )

    history_diff = ArrayField(models.CharField(max_length=200),
                              blank=True, null=True)

    data = JSONField()
    additional_data = JSONField(null=True)
    objects = HistoricalRecordQuerySet.as_manager()

    def __unicode__(self):
        return '{history_type} {content_type} id={object_id}'.format(
            history_type=self.get_history_type_display(),
            content_type=self.content_type.model,
            object_id=self.object_id
        )

    class Meta:
        app_label = 'atris'
        ordering = ['-history_date']
        abstract = True

    def get_diff_to_prev_string(self):
        """
        Generates a string which describes the changes that occurred between
        this historical record instance (self) and its previous version.

        :return: Said string.
        :rtype String
        """
        diff_string = '{}d '.format(self.get_history_type_display())

        if 'Update' not in diff_string:
            return '{action}{object}'.format(
                action=diff_string,
                object=self.content_type.model.capitalize()
            )

        if self.history_diff is None:
            previous_version = self._get_prev_version()
            if previous_version:
                self.history_diff = (
                    [self.content_type.model_class()._meta.get_field(
                        attr).verbose_name for
                     (attr, val) in self.data.items()
                     if (attr, val) not in previous_version.data.items()] or
                    ['with no change']
                )
            else:
                self.history_diff = []
            self.save()

        if not self.history_diff:
            return 'No prior information available.'

        diff_string += ', '.join(sorted(self.history_diff))

        return diff_string

    def _get_prev_version(self):
        return self.__class__.objects.previous_version_by_model_and_id(
            model=self.content_type,
            object_id=self.object_id,
            history_id=self.id
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.contrib.postgres.fields
import django.contrib.postgres.fields.hstore


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('atris', '0003_added_indexes'),
    ]

    operations = [
        migrations.CreateModel(
            name='ArchivedHistoricalRecord',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField(db_index=True)),
                ('history_date', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('history_user', models.CharField(max_length=50, null=True)),
                ('history_user_id', models.PositiveIntegerField(null=True)),
                ('history_type', models.CharField(max_length=1, choices=[('+', 'Create'), ('~', 'Update'), ('-', 'Delete')])),
                ('history_diff', django.contrib.postgres.fields.ArrayField(size=None, null=True, base_field=models.CharField(max_length=200), blank=True)),
                ('data', django.contrib.postgres.fields.hstore.HStoreField()),
                ('additional_data', django.contrib.postgres.fields.hstore.HStoreField(null=True)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
                'ordering': ['-history_date'],
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0002_historicalrecord_history_diff'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalrecord',
            name='history_date',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='object_id',
            field=models.PositiveIntegerField(db_index=True),
        ),
    ]



from atris.models import HistoricalRecord
from atris.models import history_logging
from atris.models.archived_historical_record import ArchivedHistoricalRecord
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.db import models, connections
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _


class ContentTypeListFilter(admin.SimpleListFilter):
    # Human-readable title which will be displayed in the
    # right admin sidebar just above the filter options.
    title = _('Content type')

    # Parameter for the filter that will be used in the URL query.
    parameter_name = 'content_type'

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples. The first element in each
        tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        """
        followed_content_types = [
            ContentType.objects.get_for_model(model)
            for model in history_logging.registered_models.keys()
            ]

        filter_results = set()
        for content_type in followed_content_types:
            filter_results.add((content_type.id, _(content_type.model)))

        return filter_results

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """
        if self.value() is not None:
            return queryset.filter(content_type__id=self.value())


class ApproxCountPgQuerySet(models.query.QuerySet):
    """approximate unconstrained count(*) with reltuples from pg_class"""

    def count(self):
        if self._result_cache is not None and not self._iter:
            return len(self._result_cache)

        if hasattr(connections[self.db].client.connection, 'pg_version'):
            query = self.query
            if (not query.where and query.high_mark is
                None and query.low_mark == 0 and not query.select
                and not query.group_by and not query.having
                and not query.distinct):
                parts = [p.strip('"') for p in
                         self.model._meta.db_table.split('.')]
                cursor = connections[self.db].cursor()
                if len(parts) == 1:
                    cursor.execute(
                        "select reltuples::bigint FROM pg_class WHERE relname "
                        "= %s", parts
                    )
                else:
                    cursor.execute(
                        "select reltuples::bigint FROM pg_class c JOIN "
                        "pg_namespace n on (c.relnamespace = n.oid) WHERE "
                        "n.nspname = %s AND c.relname = %s", parts
                    )
            return cursor.fetchall()[0][0]
        return self.query.get_count(using=self.db)


class GenericHistoryAdmin(admin.ModelAdmin):
    list_display = (
        'object_id', 'content_type', 'history_date', 'history_type',
        'history_user'
    )
    fields = ['object_id', 'content_type', 'history_date', 'history_type',
              'history_user', 'difference_to_previous', 'fields_that_differ',
              'history_snapshot', 'more_info']

    readonly_fields = fields

    search_fields = ('object_id',)

    list_filter = (ContentTypeListFilter, 'history_type')

    show_full_result_count = False

    def get_queryset(self, request):
        qs = super(GenericHistoryAdmin, self).get_queryset(request)
        return qs._clone(klass=ApproxCountPgQuerySet)

    def history_snapshot(self, obj):
        return self._dict_to_table(obj.data)

    def more_info(self, obj):
        return self._dict_to_table(obj.additional_data)

    def difference_to_previous(self, obj):
        return obj.get_diff_to_prev_string()

    def fields_that_differ(self, obj):
        if obj.history_diff:
            return ','.join(obj.history_diff)
        return None

    def _dict_to_table(self, dict):
        table = '<table style="border: 1px solid #eee;">' + ''.join(
            [
                '<tr><td style="border: 1px solid #eee;">{}</td> <td>{'
                '}</td></tr>'.format(
                    key, val)
                for (key, val) in dict.items()]
        ) + '</table>'

        return mark_safe(table)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class HistoricalRecordAdmin(GenericHistoryAdmin):
    model = HistoricalRecord
    show_full_result_count = False


class ArchivedHistoricalRecordAdmin(GenericHistoryAdmin):
    model = ArchivedHistoricalRecord
    show_full_result_count = False


admin.site.register(HistoricalRecord, HistoricalRecordAdmin)
admin.site.register(ArchivedHistoricalRecord, ArchivedHistoricalRecordAdmin)



import logging
from optparse import make_option
from atris.models import HistoricalRecord

from django.core.management import BaseCommand

logger = logging.getLogger('old_history_deleting')


class Command(BaseCommand):
    help = (
        """
        Deletes historical records older than the specified days or months.
        You must supply either the days or the weeks param.
        """
    )
    PARAM_ERROR = 'You must supply either the days or the weeks param'

    option_list = BaseCommand.option_list + (
        make_option('--days',
                    dest='days',
                    type='int',
                    default=None,
                    help=('Any historical record older than the number of days'
                          ' specified gets deleted.')),
        make_option('--weeks',
                    dest='weeks',
                    type='int',
                    default=None,
                    help=('Any historical record older than the number of'
                          ' months specified gets deleted.')),
    )

    def handle(self, *args, **options):
        days = options.get('days')
        weeks = options.get('weeks')
        if not (days or weeks):
            self.stderr.write("{msg}\n".format(
                msg=self.PARAM_ERROR,
            ))
            return
        old_history_entries = HistoricalRecord.objects.older_than(days, weeks)
        handled_entries_nr = old_history_entries.count()
        old_history_entries.delete()
        self.stdout.write('{} deleted.\n'.format(handled_entries_nr))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0004_archivedhistoricalrecord'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedhistoricalrecord',
            name='history_type',
            field=models.CharField(db_index=True, max_length=1, choices=[('+', 'Create'), ('~', 'Update'), ('-', 'Delete')]),
        ),
        migrations.AlterField(
            model_name='historicalrecord',
            name='history_type',
            field=models.CharField(db_index=True, max_length=1, choices=[('+', 'Create'), ('~', 'Update'), ('-', 'Delete')]),
        ),
    ]



import logging
from datetime import timedelta
from optparse import make_option

from django.core.management import BaseCommand
from django.db import transaction, connection
from django.utils.timezone import now

from atris.models import HistoricalRecord

logger = logging.getLogger('old_history_archiving')


class Command(BaseCommand):
    help = (
        """
        Archives historical records older than the specified days or months.
        You must supply either the days or the weeks param.
        The historical entries older than the specified days will be moved to
        the "atris_archivedhistoricalrecord" table.
        """
    )
    PARAM_ERROR = 'You must supply either the days or the weeks param'

    option_list = BaseCommand.option_list + (
        make_option('--days',
                    dest='days',
                    type='int',
                    default=None,
                    help=('Any historical record older than the number of days'
                          ' specified gets archived.')),
        make_option('--weeks',
                    dest='weeks',
                    type='int',
                    default=None,
                    help=('Any historical record older than the number of'
                          ' months specified gets archived.')),
    )

    def handle(self, *args, **options):
        days = options.get('days')
        weeks = options.get('weeks')
        if not (days or weeks):
            self.stderr.write("{msg}\n".format(
                msg=self.PARAM_ERROR,
            ))
            return
        old_history_entries = HistoricalRecord.objects.older_than(days, weeks)
        handled_entries_nr = old_history_entries.count()
        self.migrate_data(days, weeks)
        self.stdout.write('{} archived.\n'.format(handled_entries_nr))

    @transaction.atomic
    def migrate_data(self, days=None, weeks=None):
        if days and weeks:
            logger.info('You supplied both days and weeks, weeks param'
                        ' will be used as the delimiter.')
        td = timedelta(weeks=weeks) if weeks else timedelta(days=days)
        older_than_date = now() - td
        cursor = connection.cursor()
        fields_str = ','.join(
            [field.attname for field in HistoricalRecord._meta.fields]
        )
        query = ("INSERT INTO atris_archivedhistoricalrecord ({}) "
                 "SELECT {} FROM atris_historicalrecord "
                 "WHERE history_date < '{}';".format(fields_str, fields_str,
                                                     older_than_date.date()))
        cursor.execute(query)
        query = ("DELETE FROM atris_historicalrecord "
                 "WHERE history_date < '{}';".format(older_than_date.date()))
        cursor.execute(query)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.contrib.postgres.fields


class Migration(migrations.Migration):

    dependencies = [
        ('atris', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalrecord',
            name='history_diff',
            field=django.contrib.postgres.fields.ArrayField(size=None, null=True, base_field=models.CharField(max_length=200), blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import logging
import threading

from django.db import models
from django.utils import six

from.historical_record import HistoricalRecord

str = unicode if six.PY2 else str

registered_models = {}

logger = logging.getLogger(__name__)


# noinspection PyProtectedMember,PyAttributeOutsideInit
class HistoryLogging(object):
    thread = threading.local()

    def __init__(self, additional_data_param_name='',
                 excluded_fields_param_name='',
                 ignore_history_for_users=''):
        """
        :param additional_data_param_name: String used to determine which field
         on the object contains a dict holding any additional data.
        :type additional_data_param_name: str

        :param excluded_fields_param_name: String used to determine which field
            on the object contains a list holding the names of the fields which
            should not be tracked in the history.
        :param ignore_history_for_users: String used to determine which field
            on the object contains a dictionary holding the names of the users
            for which history should not be tracked.
            Dict should contain

        :type excluded_fields_param_name: str
        """
        self.additional_data_param_name = additional_data_param_name
        self.excluded_fields_param_name = excluded_fields_param_name
        self.ignore_history_for_users = ignore_history_for_users

    def get_history_user(self, instance):
        """Get the modifying user from the middleware."""
        try:
            if self.thread.request.user.is_authenticated():
                return self.thread.request.user
        except AttributeError:
            return getattr(instance, 'history_user', None)

    def contribute_to_class(self, cls, name):
        self.manager_name = name
        self.module = cls.__module__
        models.signals.class_prepared.connect(self.finalize, sender=cls)

    def finalize(self, sender, **kwargs):
        if sender not in registered_models:
            registered_models[sender] = {
                'additional_data_param_name': self.additional_data_param_name,
                'excluded_fields_param_name': self.excluded_fields_param_name
            }
        # The HistoricalRecords object will be discarded,
        # so the signal handlers can't use weak references.
        models.signals.post_save.connect(self.post_save, sender=sender,
                                         weak=False)
        models.signals.post_delete.connect(self.post_delete, sender=sender,
                                           weak=False)
        setattr(sender, self.manager_name, HistoryManager())

    def post_save(self, instance, created, **kwargs):
        if not kwargs.get('raw', False):
            self.create_historical_record(instance, created and '+' or '~')

    def post_delete(self, instance, **kwargs):
        self.create_historical_record(instance, '-')

    def create_historical_record(self, instance, history_type):
        history_user, history_user_id = self._get_user_info(instance)
        if self.skip_history_by_user(instance, history_user, history_user_id):
            logger.info("Skipping history instance for user '{}' with user id"
                        " '{}'".format(history_user, history_user_id))
            return

        data = self._get_fields_from_instance(instance)
        if history_type == '~':
            diff_fields = self._get_diff_fields(data, instance)
        else:
            diff_fields = list()

        additional_data = dict(
            (key, str(value)) for (key, value)
            in getattr(instance, self.additional_data_param_name, {}).items()
        )

        HistoricalRecord.objects.create(
            content_object=instance,
            history_type=history_type,
            history_user=history_user,
            history_user_id=history_user_id,
            data=data,
            history_diff=diff_fields,
            additional_data=additional_data
        )

    def _get_diff_fields(self, data, instance):
        if instance.history.exists():
            previous_snapshot = instance.history.first()
            return (
                [instance._meta.get_field(attr).verbose_name
                 for (attr, val) in data.items()
                 if (attr, val) not in previous_snapshot.data.items()] or
                ['with no change']
            )
        else:
            return list()

    def skip_history_by_user(self, instance, user, user_id):
        skip_dict = getattr(instance, self.ignore_history_for_users, {})
        ids_with_skip = skip_dict.get('user_ids')
        user_names_to_skip = skip_dict.get('user_names')
        if skip_dict and (
                (user_names_to_skip and (user in user_names_to_skip)) or
                (ids_with_skip and (user_id in ids_with_skip))
        ):
            return True
        return False

    def _get_fields_from_instance(self, instance):
        sentinel = object()
        data = {}
        excluded_fields = getattr(
            instance, self.excluded_fields_param_name, [])
        for field in instance._meta.fields:
            if field.attname not in excluded_fields:
                key = field.attname
                value = getattr(instance, field.attname, sentinel)
                if value is not None and value is not sentinel:
                    value = str(value)
                elif value is sentinel:
                    logger.error(('Field "{}" is invalid.'.format(key)))
                data[key] = value
        return data

    def _get_user_info(self, instance):
        user = self.get_history_user(instance)
        full_name = (user.get_full_name() if callable(
            getattr(user, 'get_full_name', None)) else None)
        username = (user.get_username() if callable(
            getattr(user, 'get_username', None)) else None)
        history_user = (
            full_name or getattr(user, 'email', None) or username
            if user else None
        )
        history_user_id = user.id if user else None
        return history_user, history_user_id


class HistoryManager(object):
    def __get__(self, instance, model):
        if instance and model:
            return HistoricalRecord.objects.by_model_and_model_id(model,
                                                                  instance.id)
        if model:
            return HistoricalRecord.objects.by_model(model)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
# Prevent `/` to return a truncated response. Can be removed in Py3
from __future__ import division  # noqa

from math import ceil
from optparse import make_option

from django.utils import six
from django.core.management import BaseCommand
from django.db import DatabaseError
from django.db.transaction import atomic

from atris.models import HistoricalRecord, registered_models

str = str if six.PY2 else str


class Command(BaseCommand):
    EXISTING_HISTORY_FOUND = 'Existing history found, skipping model'

    SELECT_BATCH_SIZE = 1000
    CREATE_BATCH_SIZE = 1000

    option_list = BaseCommand.option_list + (
        make_option('--select-batch-size',
                    dest='select_batch_size',
                    type='int',
                    default=SELECT_BATCH_SIZE,
                    help='The number of target objects that will be retrieved '
                         'from the DB at one time.'),
        make_option('--create-batch-size',
                    dest='create_batch_size',
                    type='int',
                    default=CREATE_BATCH_SIZE,
                    help='The number of history objects that will be created '
                         'in one batch. Should be at most SELECT_BATCH_SIZE.')
    )

    def handle(self, *args, **options):
        for model, model_specific_info in registered_models.items():
            if HistoricalRecord.objects.by_model(model).exists():
                self.stderr.write('{msg} {model}\n'.format(
                    msg=self.EXISTING_HISTORY_FOUND,
                    model=model,
                ))
                continue
            self.stdout.write('Initializing history for {}\n'.format(model))
            additional_data_param_name = model_specific_info[
                'additional_data_param_name']
            additional_data_field = getattr(model, additional_data_param_name,
                                            {})
            excluded_fields_param_name = model_specific_info[
                'excluded_fields_param_name']
            excluded_fields = getattr(model, excluded_fields_param_name, [])
            create_history_for_model = ModelHistoryCreator(
                model, additional_data_field, excluded_fields,
                options['select_batch_size'], options['create_batch_size'],
                self.stdout
            )
            try:
                with atomic():
                    create_history_for_model()
            except DatabaseError as e:
                self.stderr.write('Error creating history for {}: {}'
                                  .format(model, e))


class ModelHistoryCreator(object):

    def __init__(self, model, additional_data_field, excluded_fields,
                 select_batch_size, create_batch_size, output):
        self.model = model
        self.additional_data_field = additional_data_field
        self.excluded_fields = excluded_fields
        self.select_batch_size = select_batch_size
        self.create_batch_size = create_batch_size
        self.output = output

    def __call__(self):
        objects = self.model.objects
        number_of_batches = int(ceil(objects.count() / self.select_batch_size))
        self.output.write(
            'Processing data in {} batches of {} target objects.\n'
            .format(number_of_batches, self.select_batch_size)
        )
        for multiplicity in range(number_of_batches):
            start = self.select_batch_size * multiplicity
            end = start + self.select_batch_size
            self.create_history_for_objects(objects.all()[start:end])
            self.output.write('Finished batch #{} of {}.\n'
                              .format(multiplicity + 1, number_of_batches))

    def create_history_for_objects(self, objects):
        historical_instances = []
        for instance in objects:
            historical_record = self.create_history_for_object(instance)
            historical_instances.append(historical_record)
        HistoricalRecord.objects.bulk_create(historical_instances,
                                             batch_size=self.create_batch_size)

    def create_history_for_object(self, obj):
        data = self.serialize_object_data(obj)
        additional_data = {
            key: str(value)
            for key, value in self.additional_data_field.items()
        }
        historical_record = HistoricalRecord(
            history_user=None,
            history_type='+',
            content_object=obj,
            data=data,
            additional_data=additional_data
        )
        return historical_record

    def serialize_object_data(self, instance):
        data = {}
        for field in instance._meta.fields:
            if field.attname not in self.excluded_fields:
                key = field.attname
                try:
                    value = getattr(instance, field.attname)
                except AttributeError:
                    self.output.write(('Field "{}" is invalid.'.format(key)))
                else:
                    if value is not None:
                        value = str(value)
                        data[key] = value
        return data
